/*
 * Copyright (c) 2002 Michal Moskal <malekith@pld-linux.org>
 * All rights reserved. See file COPYRIGHT for details about licensing.
 *
 * $Rev: 946 $ $LastChangedDate: 2002-09-12 22:50:25 +0200 (Thu, 12 Sep 2002) $
 *
 * Lex - The lexer. 
 *
 * It is meant to be used with bison using %locations and thread safe
 * interface to lexing function (location interface isn't thread safe
 * though, ask bison people).
 */

#include <unistd.h>
#include <string.h>
#include <stdlib.h>
#include <stdio.h>
#include <fcntl.h>

#include <gont/runtime.h>

#include "parser.h"
#include "cgont/ast.h"
#include "cgont/util.h"

YYLTYPE last_lex_loc;

extern YYSTYPE yylval;
extern YYLTYPE yylloc;
#define lvalp (&yylval)
#define llocp (&yylloc)

static struct lexer_state *lex;

struct keyword { char *name; int tok; };
struct keyword *lookup_keyword(const char*, unsigned int);

/* This buffering is easy transferable to mmap() style. */
struct buffer {
	char *data;
	int data_sz, data_len, data_ptr;
	
	const char *line_ptr;
	int line_len;

	int line_no;

	int fd;
	int eof;
};

enum basic_token_kind {
	btk_id,
	btk_op,
	btk_int,
	btk_float,
	btk_string,
	btk_char,
	btk_tyvar,
	btk_error,
	btk_eof,
};

struct basic_token {
	int beg, end;
	enum basic_token_kind kind;
	long data;
	double dbl;
	struct gont_string *string;
};

struct lexer_state {
	struct buffer *buf;
	struct basic_token tok;
	int pos;
	struct gont_string *filename;
};

/* Needed by c_parser.y.  */
struct GontS_Ast__location *make_location(YYLTYPE *l)
{
	struct GontS_Ast__location *loc = gont_xmalloc(sizeof(*loc));
	loc->file = lex->filename;
	loc->first_line = l->first_line;
	loc->last_line = l->last_line;
	loc->first_column = l->first_column;
	loc->last_column = l->last_column;
	return loc;
}

/* Likewise.  */
void yyerror(const char *msg)
{
	Util$$errorl(make_location(&last_lex_loc), gont_string_utf8(msg));
}

static int buf_read_more_data(struct buffer *buf)
{
	int data_left = buf->data_len - buf->data_ptr;
	int nread;
	
	if (data_left * 2 >= buf->data_sz) {
		char *p;
		if (data_left == 0)
			buf->data_sz = 16;
		else
			buf->data_sz = data_left * 4;
		p = gont_xmalloc_atomic(buf->data_sz);
		memcpy(p, buf->data + buf->data_ptr, data_left);
		gont_free(p);
		buf->data = p;
	} else {
		memmove(buf->data, buf->data + buf->data_ptr, data_left);
	}

	buf->data_ptr = 0;
	buf->data_len = data_left;
	
	nread = read(buf->fd, buf->data + data_left, buf->data_sz - data_left);
	
	if (nread <= 0)
		return -1;
	
	buf->data_len += nread;

	return 0;
}

static int buf_corner_case(struct buffer *buf)
{
	int len = buf->data_len;
	
	buf->eof++;

	if (buf->data_len == 0) {
		buf->line_len = 0;
		return -1;
	}
	
	if (buf->data[len - 1] == '\r')
		len--;
	
	buf->line_ptr = buf->data;
	buf->line_len = len;

	return 0;
}

static int buf_readline(struct buffer *buf)
{
	int i;
	
	if (buf->eof) {
		buf->line_len = 0;
		return -1;
	}
	
	buf->line_no++;
	
	for (;;) {
		for (i = buf->data_ptr; i < buf->data_len; i++)
			if (buf->data[i] == '\n' || buf->data[i] == '\r')
				break;
				
		if (i == buf->data_len || 
		    (i == buf->data_len - 1 && buf->data[i] == '\r')) {
			if (buf_read_more_data(buf))
				return buf_corner_case(buf);
		} else
			break;
	}

	buf->line_ptr = buf->data + buf->data_ptr;
	buf->line_len = i - buf->data_ptr;
	
	if (buf->data[i] == '\r' && buf->data[i + 1] == '\n')
		i++;
	buf->data_ptr = i + 1;

	return 0;
}

static struct buffer *buf_new(int fd)
{
	struct buffer *buf;

	buf = gont_xmalloc0(sizeof(struct buffer));
	buf->fd = fd;

	return buf;
}

static void eat_comment(struct lexer_state *lex)
{
	int p = ++lex->pos;
	int level = 1;
	int start_line = lex->buf->line_no;
	int start_pos = lex->pos - 2;
	
	for (;;) {
		const char *line = lex->buf->line_ptr;
		int len = lex->buf->line_len - 1;
		
		while (p < len) {
			if (line[p] == '*' && line[p + 1] == '/') {
				if (--level == 0)
					break;
			} else if (line[p] == '/' && line[p + 1] == '*') {
				level++;
			}
			p++;
		}

		if (level == 0) {
			lex->pos = p + 2;
			return;
		}

		if (buf_readline(lex->buf)) {
			last_lex_loc.first_line = start_line;
			last_lex_loc.first_column = start_pos;
			last_lex_loc.last_line = lex->buf->line_no;
			last_lex_loc.last_column = p;
			yyerror("runaway comment");
			lex->tok.kind = btk_error;
			return;
		}

		p = 0;
	}
}

static void lex_error(struct lexer_state *lex, const char *msg)
{
	last_lex_loc.first_line = lex->buf->line_no;
	last_lex_loc.first_column = lex->pos;
	last_lex_loc.last_line = 0;
	last_lex_loc.last_column = 0;
	yyerror(msg);
	lex->tok.kind = btk_error;
}

static int lex_escape(struct lexer_state *lex, int *p)
{
	(*p)++;
	if (*p < lex->buf->line_len) {
		switch (lex->buf->line_ptr[*p]) {
		case 'n':
			return '\n';

		case 'r':
			return '\r';
			
		case 't':
			return '\t';
			
		case 'b':
			return '\b';

		case 'e':
			return '\e';

		case '"':
		case '\'':
		case '\\':
			return lex->buf->line_ptr[*p];

		case 'x':
		case 'X':
			if (*p + 2 < lex->buf->line_len) {
				char buf[3], *pp;
				int r;
				
				buf[0] = lex->buf->line_ptr[++*p];
				buf[1] = lex->buf->line_ptr[++*p];
				buf[2] = 0;
				pp = buf;
				r = strtol(buf, &pp, 16);
				if (*pp != 0) {
					lex_error(lex, "invalid hex escape");
					r = -1;
				}

				return r;
			} else {
				lex_error(lex, "truncated hex escape");
				return -1;
			}
	
		default:
			lex_error(lex, "invalid string escape");
			return -1;
		}
	} else {
		lex_error(lex, "truncated string escape");
		return -1;
	}
}

static void lex_string(struct lexer_state *lex)
{
	int p = lex->pos;
	const char *line = lex->buf->line_ptr;
	unsigned char buf[1024], *ptr;
	int len = 0;
	
	lex->tok.kind = btk_string;

	while (p < lex->buf->line_len) {
		if (line[p] == '\\')
			lex_escape(lex, &p);
		else if (line[p] == '"')
			break;
		len++;
		p++;
	}

	if (lex->tok.kind == btk_error)
		return;
		
	if (p >= lex->buf->line_len) {
		lex_error(lex, "runaway string");
		return;
	}

	if (len < (int)sizeof(buf))
		ptr = buf;
	else
		ptr = gont_xmalloc_atomic(len);

	len = 0;
	p = lex->pos;
	
	while (p < lex->buf->line_len) {
		if (line[p] == '\\') {
			ptr[len++] = lex_escape(lex, &p);
			p++;
		} else if (line[p] == '"') {
			break;
		} else {
			ptr[len++] = line[p++];
		}
	}

	lex->pos = p + 1;
	lex->tok.end = p;

	lex->tok.string = Util$$make_string(make_location(&last_lex_loc), 
					    ptr, len);
}

static void lex_id(struct lexer_state *lex)
{
	int p = lex->pos, ch;

	while (p < lex->buf->line_len) {
		ch = lex->buf->line_ptr[p++];
		
		if (!((ch >= 'a' && ch <= 'z') ||
		      (ch >= 'A' && ch <= 'Z') ||
		      (ch >= '0' && ch <= '9') ||
		       ch == '_' || ch == '\'')) {
		       	p--;
			break;
		}
	}

	lex->pos = p;
	lex->tok.end = p - 1;
	lex->tok.kind = btk_id;
}

static void lex_number(struct lexer_state *lex)
{
	int p = lex->pos, ch;
	int seen_dot = 0, seen_x = 0, seen_e = 0;

	while (p < lex->buf->line_len) {
		ch = lex->buf->line_ptr[p++];
		
		if (ch == 'x')
			seen_x++;

		if (ch == 'e')
			seen_e++;

		if (ch == '.')
			seen_dot++;

		if (!((ch >= 'a' && ch <= 'z') ||
		      (ch >= 'A' && ch <= 'Z') ||
		      (ch >= '0' && ch <= '9') ||
		       ch == '_' || ch == '.')) {
		        p--;
			break;
		}
	}

	lex->pos = p;
	lex->tok.end = p - 1;

	if (seen_dot || (seen_e && !seen_x))
		lex->tok.kind = btk_float;
	else
		lex->tok.kind = btk_int;
}

static void lex_tyvar_or_char(struct lexer_state *lex)
{
	int p = lex->pos;
	
	if (p == lex->buf->line_len) {
		lex_error(lex, "single '");
		return;
	}
		
	if (lex->buf->line_ptr[p] == '\\') {
		lex->tok.data = lex_escape(lex, &p);
		if (++p >= lex->buf->line_len || 
		    lex->buf->line_ptr[p] != '\'') {
			lex_error(lex, "unclosed character literal");
			return;
		}
		lex->pos = p + 1;
		lex->tok.end = p;
		lex->tok.kind = btk_char;
	} else if (p + 1 < lex->buf->line_len && 
		   lex->buf->line_ptr[p + 1] == '\'') {
		lex->tok.data = (unsigned char)lex->buf->line_ptr[p];
		lex->pos = ++p + 1;
		lex->tok.end = p;
		lex->tok.kind = btk_char;
	} else {
		int ch = (unsigned char)lex->buf->line_ptr[p];
		if ((ch >= 'a' && ch <= 'z') ||
		    (ch >= 'A' && ch <= 'Z') ||
		    ch == '_') {
		        lex->pos = p + 1;
		    	lex_id(lex);
			lex->tok.kind = btk_tyvar;
		} else {
			lex_error(lex, "invalid character in type variable");
		}
	}
}

static void lex_read_basic_token(struct lexer_state *lex)
{
	int p = lex->pos, ch = 0, ch2, ch3;

	for (;;) {
		while (p < lex->buf->line_len) {
			ch = lex->buf->line_ptr[p++];
			if (ch != ' ' && ch != '\t')
				goto out;
		}

		if (buf_readline(lex->buf)) {
			lex->tok.kind = btk_eof;
			return;
		}
		p = 0;
	}

out:

	lex->pos = p;
	lex->tok.beg = --p;
	
	lex->tok.kind = btk_op;
	lex->tok.data = ch;
	lex->tok.end = lex->pos;
	
	ch2 = lex->pos >= lex->buf->line_len ? 
				-1 : lex->buf->line_ptr[lex->pos];
	ch3 = lex->pos >= lex->buf->line_len - 1 ? 
				-1 : lex->buf->line_ptr[lex->pos + 1];

#define CH2(second, token) 			\
	if (ch2 == second) { 			\
		lex->tok.end++;			\
		lex->pos++;			\
		lex->tok.data = token;		\
		break;				\
	}
		
#define CH3(second, third, token)		\
	if (ch2 == second && ch3 == third) {	\
		lex->tok.end += 2;		\
		lex->pos += 2;			\
		lex->tok.data = token;		\
		break;				\
	}
		
	switch (ch) {
	case '/':
		switch (ch2) {
		case '=':
			lex->tok.end++;
			lex->pos++;
			lex->tok.data = SLASH_EQUAL;
			break;
		case '*':
			eat_comment(lex);
			lex_read_basic_token(lex);
			break;
		case '/':
			lex->pos = lex->buf->line_len;
			lex_read_basic_token(lex);
			break;
		}
		break;
		
	case '+':
		CH2('+', PLUS_PLUS);
		CH2('=', PLUS_EQUAL);
		break;
		
	case '-':
		CH2('=', MINUS_EQUAL);
		CH2('-', MINUS_MINUS);
		CH2('>', ARROW);
		break;
		
	case '<':
		CH2('=', LE);
		CH3('<', '=', LTLT_EQUAL);
		CH2('<', LTLT);
		break;
		
	case '>':
		CH2('=', GE);
		CH3('>', '=', GTGT_EQUAL);
		CH2('>', GTGT);
		break;
		
	case '=':
		CH2('=', EQ);
		break;
		
	case '!':
		CH2('=', NE);
		break;
		
	case ':':
		CH2(':', DOUBLECOLON);
		break;
		
	case '%':
		CH2('=', PERCENT_EQUAL);
		break;
		
	case '*':
		CH2('=', STAR_EQUAL);
		break;
		
	case '.':
	case '@':
	case '(':
	case ')':
	case '[':
	case ']':
	case '{':
	case '}':
	case ';':
	case ',':
		break;
	
	case '"':
		lex_string(lex);
		break;
	
	case '\'':
		lex_tyvar_or_char(lex);
		break;
	
	case '&':
		CH2('&', AND_AND);
		CH2('=', AND_EQUAL);
		break;
		
	case '|':
		CH2('|', OR_OR);
		CH2('=', OR_EQUAL);
		break;
		
	case '^':
		CH2('=', XOR_EQUAL);
		break;
		
	default:
		if ((ch >= 'a' && ch <= 'z') ||
		    (ch >= 'A' && ch <= 'Z') ||
		    ch == '_') {
		    	lex_id(lex);
			if (lex->tok.beg == lex->tok.end &&
			    lex->buf->line_ptr[lex->tok.beg] == '_') {
				lex->tok.kind = btk_op;
			}
		} else if (ch >= '0' && ch <= '9')
			lex_number(lex);
		else {
			lex_error(lex, "invalid character in input");
		}
		break;
	}
}

int yylex(void)
{
	struct keyword *kw;
	const char *p;
	int len;
	
	lex_read_basic_token(lex);

	llocp->first_line = llocp->last_line = lex->buf->line_no;
	llocp->first_column = lex->tok.beg;
	llocp->last_column = lex->tok.end;

	last_lex_loc = *llocp;

	p = lex->buf->line_ptr + lex->tok.beg;
	len = lex->tok.end - lex->tok.beg + 1;

	switch (lex->tok.kind) {
	case btk_id:
		kw = lookup_keyword(p, len);
		if (kw) {
			return kw->tok;
		} else {
			*lvalp = gont_string_of_utf8(p, len);
			return ID;
		}
		
	case btk_tyvar:
		*lvalp = gont_string_of_utf8(p + 1, len - 1);
		return TYVAR;
	
	case btk_op:
		return lex->tok.data;
	
	case btk_string:
		*lvalp = lex->tok.string;
		lex->tok.string = NULL;
		return STRING;
	
	case btk_int:
		*lvalp = gont_string_of_utf8(p, len);
		return INT;
		
	case btk_float:
		*lvalp = gont_string_of_utf8(p, len);
		return FLOAT;

	case btk_char:
		*lvalp = (void*)lex->tok.data;
		return CHAR;
	
	case btk_error:
		*lvalp = NULL;
		return LEX_ERROR;

	case btk_eof:
		return 0;
		
	default:
		abort();
	}
}

void make_lexer(struct gont_string *filename)
{
	int fd;
	GONT_STRING_TO_CSTRING_N(filename, gont_default_charset);

	fd = open(c_filename, O_RDONLY);
	if (fd < 0) {
		perror(c_filename);
		exit(1);
	}

	lex = gont_xmalloc0(sizeof(*lex));
	lex->buf = buf_new(fd);
	lex->filename = filename;
}
