/* 
 * Copyright (c) 2002 Michal Moskal <malekith@pld-linux.org>
 * All rights reserved. See file COPYRIGHT for details about licensing.
 *
 * $Rev: 931 $ $LastChangedDate: 2002-09-11 10:32:30 +0200 (Wed, 11 Sep 2002) $
 *
 * Standard library for Gont, Bitvector module C implementation.
 */

#include <gont/runtime.h>

#include <assert.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>

struct gont_bitvector {
	gont_int size;
	gont_int *data;
};

typedef struct gont_bitvector *t;

#define BITS_TO_WORDS(sz) ((sz + GONT_WORD_SIZE - 1) / GONT_WORD_SIZE)
#define BITS_TO_BYTES(sz) (BITS_TO_WORDS(sz) * sizeof(gont_int))
#define SET(ar, bit) \
	(ar)[(bit) / GONT_WORD_SIZE] |= 1 << (bit) % GONT_WORD_SIZE
#define CLR(ar, bit) \
	(ar)[(bit) / GONT_WORD_SIZE] &= ~(1 << (bit) % GONT_WORD_SIZE)
#define GET(ar, bit) \
	((ar)[(bit) / GONT_WORD_SIZE] & 1 << (bit) % GONT_WORD_SIZE)

/* [sz] is number of bits.  */
t gont_make_bitvector(gont_int sz)
{
	t ret = gont_xmalloc(sizeof(*ret));

	ret->data = gont_xmalloc_atomic(BITS_TO_BYTES(sz));
	ret->size = sz;

	return ret;
}

static void zero_pad(t a)
{
	int last_sz = a->size % GONT_WORD_SIZE;

	if (last_sz) {
		gont_int mask = 0;
		mask = ~(~mask << last_sz);
		a->data[a->size / GONT_WORD_SIZE] &= mask;
	}
}

t Bitvector__make(gont_int sz, gont_int ini)
{
	t ret = gont_make_bitvector(sz);

	if (ini) {
		memset(ret->data, 0xff, BITS_TO_BYTES(sz));
		zero_pad(ret);
	} else {
		memset(ret->data, 0x00, BITS_TO_BYTES(sz));
	}
	
	return ret;
}

t Bitvector__copy(t src)
{
	t ret = gont_make_bitvector(src->size);

	memcpy(ret->data, src->data, BITS_TO_BYTES(src->size));

	return ret;
}

t Bitvector__init(gont_int sz, struct gont_lambda_expr *ini)
{
	t ret = gont_make_bitvector(sz);
	gont_int i, j, k;

	memset(ret->data, 0x00, BITS_TO_BYTES(sz));
	
	i = j = k = 0;
	
	for (i = 0; i < sz; k++, i++) {
		if (k == GONT_WORD_SIZE) {
			k = 0;
			j++;
		}
		
		if (ini->fnc(ini->closure, i))
			ret->data[j] |= 1 << k;
	}

	return ret;
}

gont_int Bitvector__length(t ar)
{
	return ar->size;
}

void Bitvector__set(t ar, gont_int idx, gont_int val)
{
	if (idx >= ar->size)
		gont_invalid_argument("Bitvector::set");

	if (val)
		SET(ar->data, idx);
	else
		CLR(ar->data, idx);
}

gont_int Bitvector__get(t ar, gont_int idx)
{
	if (idx >= ar->size)
		gont_invalid_argument("Bitvector::get");
		
	if (GET(ar->data, idx))
		return 1;
	else
		return 0;
}

void Bitvector__resize(t ar, gont_int newsz, gont_int ini)
{
	gont_int *newdata = gont_xmalloc_atomic(BITS_TO_BYTES(newsz));

	if (newsz > ar->size) {
		gont_int p;
		
		memcpy(newdata, ar->data, BITS_TO_BYTES(ar->size));
		
		p = ar->size;
		
		while (p % GONT_WORD_SIZE) {
			if (ini)
				SET(newdata, p);
			else
				CLR(newdata, p);
			p++;
		}

		if (ini)
			ini = 0xff;
			
		memset(newdata + BITS_TO_WORDS(p), ini, 
		       BITS_TO_BYTES(newsz) - BITS_TO_BYTES(p));
	} else {
		memcpy(newdata, ar->data, BITS_TO_BYTES(newsz));
	}

	ar->size = newsz;
	ar->data = newdata;
	
	zero_pad(ar);
}

void Bitvector___op(t dst, t src, int op)
{
	gont_int i, n;
	
	if (dst->size < src->size)
		Bitvector__resize(dst, src->size, 0);
	
	n = BITS_TO_WORDS(src->size);

	switch (op) {
	case 0:
		for (i = 0; i < n; i++)
			dst->data[i] |= src->data[i];
		break;
		
	case 1:
		for (i = 0; i < n; i++)
			dst->data[i] &= src->data[i];
		memset(dst->data + n, 0x00,
		       BITS_TO_BYTES(dst->size) - BITS_TO_BYTES(src->size));
		break;
		
	case 2:
		for (i = 0; i < n; i++)
			dst->data[i] ^= src->data[i];
		break;
		
	default:
		gont_oops("Bitvector___op: bad op");
	}
}
