/*
 * Copyright (c) 2002 Michal Moskal <malekith@pld-linux.org>
 * All rights reserved. See file COPYRIGHT for details about licensing.
 *
 * $Rev: 934 $ $LastChangedDate: 2002-09-11 18:22:59 +0200 (Wed, 11 Sep 2002) $
 *
 * Cg_expr - code generation for expressions.
 */

open Ast;
open Util;
open Tyutil;

open Cg_util;
open Cg_types;
open Cg_globals;
open Cg_expr;
open Cg_stmt;

string parent_closure_type_name(cg_ctx c)
{
    switch c.parent_func {
    case Some(f):
	return f.cf_closure_type_name;
    case None:
	return "";
    }
}

Ksi::expr local_ref(var_symbol s)
{
    let r = Ksi::Ref(make_id(s.name)) in
    if (has_property(s, Is_in_closure))
	return Ksi::Indirect(r);
    else
	return r;
}

Ksi::expr compile_ref(cg_ctx c, ref_record rr)
{
    let s = get_binding(rr.ref_symbol) in
    if (s.a.fnc_level <= 0) {
	return global_ref(s);
    } else if (rr.ref_level == 0) {
	return local_ref(s);
    } else {
	if (!Ptrhash::mem(c.cf_references_hash, s)) {
	    Ksi::expr go_up(int n)
	    {
		if (n == 1)
		    return Ksi::Ref(parent_closure_id);
		else
		    return Ksi::Arrow(go_up(n - 1), parent_closure_id);
	    }
	  
	    let in_parent = Ksi::Arrow(go_up(rr.ref_level),
				       make_id(s.a.extern_name)) in
	    let vi = make_vardef(Ksi::Reference(storage_type(s.stype)), 
				 make_id(s.name), 
				 raddr(in_parent)) in {
		Ptrhash::add(c.cf_references_hash, s, 1);
		ref_set(c.cf_references, Cons(vi, ref_get(c.cf_references)));
	    }
	}

	return Ksi::Indirect(Ksi::Ref(make_id(s.name)));
    }
}

Ksi::expr vardef(cg_ctx, var_symbol s)
{
    let t = storage_type(s.stype) in
    let n = make_id(s.name) in
    if (has_property(s, Is_in_closure)) {
	let in_closure = Ksi::Arrow(Ksi::Ref(our_closure_id), 
				    make_id(s.a.extern_name)) in
	return make_vardef(Ksi::Reference(t), n, raddr(in_closure));
    } else {
	return Ksi::Var(t, n);
    }
}

Ksi::expr vardef_init(cg_ctx c, var_symbol s, Ksi::expr ini)
{
    return pair(vardef(c, s), Ksi::Modify(local_ref(s), ini));
}

void push_closure_type(cg_ctx c, <var_symbol>list closure_vars)
{
    int cnt = 1;
    
    *(Ksi::ty, Ksi::id) compile_cv(var_symbol s)
    {
	s.a.extern_name = s.name + "$" + itoa(cnt++);
	return (storage_type(s.stype), make_id(s.a.extern_name));
    }
   
//dbg("fnc: " + c.cf_name + ": " + itoa(List::length(closure_vars)) + "\n");

    <*(Ksi::ty, Ksi::id)>list vars = 
		 List::map(compile_cv, closure_vars);

    let p = parent_closure_type_name(c) in
    if (p != "") {
	let fld = (Ksi::Ptr(Ksi::Struct(make_id(p))), parent_closure_id) in
	vars = Cons(fld, vars);
    }
	
    if (List::length(vars) == 0) {
	ice("Codegen::push_closure_type(): needs closure but has none");
    } else {
	c.cf_closure_type_name = c.cf_name + "$closure_struct";
	let id = make_id(c.cf_closure_type_name) in
	output_decl(Ksi::Struct_def(Nil, id, vars));
    }
}

Ksi::binop ksi_op2(op2 o)
{
    switch o {
    case Plus:	    return Ksi::Plus;
    case Minus:	    return Ksi::Minus;
    case Mult:	    return Ksi::Mult;
    case Mod:	    return Ksi::Trunc_mod;
    case Div:	    return Ksi::Trunc_div;
    case Eq:	    return Ksi::Eq;
    case Ne:	    return Ksi::Ne;
    case Lt:	    return Ksi::Lt;
    case Gt:	    return Ksi::Gt;
    case Ge:	    return Ksi::Ge;
    case Le:	    return Ksi::Le;
    case And_if:    return Ksi::Truth_and_if;
    case Or_if:	    return Ksi::Truth_or_if;
    case Bit_and:   return Ksi::Bit_and;
    case Bit_or:    return Ksi::Bit_ior;
    case Bit_xor:   return Ksi::Bit_xor;
    case Bit_lshift:return Ksi::L_shift;
    case Bit_rshift:return Ksi::R_shift;
    }
}

// opt-TODO: we can skip saving of last expression with side-effects
*(<Ksi::expr>list, <Ksi::expr>option) call_args(cg_ctx c, <expr>list args,
						   <ty>list formal_types)
{
    Ksi::expr ini = nop;
    int pos = 0;
    
    Ksi::expr map_arg(expr e)
    {
	Ksi::expr x = compile_expr(c, e);
	
	if (has_side_effects(e)) {
	    let t = storage_type(type_of(e)) in
	    let n = make_id("arg" + itoa(pos++) + "$") in {
		ini = pair(ini, make_vardef(t, n, x));
		x = Ksi::Ref(n);
	    }
	}

	x = s2t_conversion(type_of(e), x);

	switch formal_types {
	case Cons(hd, tl):
	    formal_types = tl;
	    if (is_poly(hd))
		x = Ksi::Convert(transport_type(hd), x);
		
	case Nil:
	    skip;
	}

	return x;
    }
   
    let ret = List::map(map_arg, args) in
    switch ini {
    case Ksi::Nop:
	return (ret, None);
    case _:
	return (ret, Some(ini));
    }
}

Ksi::expr maybe_save(cg_ctx c, expr to_save, *(Ksi::expr (Ksi::expr)) in_what)
{
    let compiled = compile_expr(c, to_save) in
    switch to_save.a.raw {
    case Ref(_):
	// ok, we don't bother saving
	return in_what(compiled);
	
    case _:
	// but here we should, or even have to (if expression has
	// side effects)
	let n = make_id("saved$") in
	let t = storage_type(type_of(to_save)) in
	let save = make_vardef(t, n, compiled) in
	return Ksi::Bind(pair(save, in_what(Ksi::Ref(n))));
    }
}

Ksi::expr compile_call(cg_ctx c, ty our_type, expr fnc, <expr>list args)
{
    <Ksi::expr>option ini;
    <Ksi::expr>list cargs;

    let formal_types = List::map(fst, (get_ft_record(instanced_type_of(fnc))).parms) in
    (cargs, ini) = call_args(c, args, formal_types);
    
    Ksi::expr lambda_call()
    {
	let ft = get_ft_record(instanced_type_of(fnc)) in
	let targs = List::map(transport_type, List::map(fst, ft.parms)) in
	let cl_args = Cons(Ksi::Poly, targs) in {
	    Ksi::ty fnc_type = Ksi::Func_ptr(transport_type(ft.ret_type), 
					    cl_args);

	    Ksi::expr make_call(Ksi::expr ptr)
	    {
		let fnc_ptr = Ksi::Arrow(ptr, make_id("func_ptr")) in
		let closure = Ksi::Arrow(ptr, make_id("closure")) in
		let allargs = Cons(closure, cargs) in
		return Ksi::Ptr_call(Ksi::Convert(fnc_type, fnc_ptr), allargs);
	    }

	    return maybe_save(c, fnc, make_call);
	}
    }
    
    Ksi::expr call;
    
    // optimize calls to global functions
    switch fnc.a.raw {
    case Ref(rr):
	let s = get_binding(rr.ref_symbol) in
	if (is_global_function(s)) {
	    push_proto(s);
	    call = Ksi::Call(make_id(s.a.extern_name), cargs);
	} else {
	    call = lambda_call();
	}
    case _:
	call = lambda_call();
    }

    call = t2s_conversion(our_type, call);

    switch type_instance(our_type) {
    case Void:
	call = pair(call, zero_const);
    case _:
	skip;
    }

    switch ini {
    case Some(i):
	call = Ksi::Bind(pair(i, call));
    case None:
	skip;
    }

    if (is_poly(type_of(fnc)))
	call = Ksi::Convert(storage_type(our_type), call);

    return call;
}

Ksi::expr compile_tuple(cg_ctx c, <expr>list args)
{
    let (a, i) = call_args(c, args, Nil) in
    let cnt = Ksi::Int_const(ptroff, Int64::of_int(List::length(args))) in
    let call = Ksi::Call(make_id("gont_make_tuple"), Cons(cnt, a)) in {
	switch i {
	case Some(ini):
	    return Ksi::Bind(pair(ini, call));
	    
	case None:
	    return call;
	}
    }
}

Ksi::expr compile_struct_expr(cg_ctx c, ty our_type, <sini_expr>list inis,
			      <Ksi::expr>option copy_from)
{
    Ksi::expr sz = size_of_struct(our_type);
    
    let n = make_id("sx$") in
    let alloc = Ksi::Convert(storage_type(our_type), gont_alloc(sz)) in
    let v = make_vardef(storage_type(our_type), n, alloc) in {
	Ksi::expr ref = Ksi::Ref(n);
	Ksi::expr ini = v;

	switch copy_from {
	case Some(ex):
	    ini = pair(ini, Ksi::Built_in(make_id("memcpy"), [ref, ex, sz]));
	    
	case None:
	    skip;
	}
	
	void init_one(sini_expr se)
	{
	    let s = get_binding(se.sx_symbol) in
	    let x = compile_expr(c, se.sx_expr) in
	    let i = Ksi::Modify(Ksi::Arrow(ref, make_id(s.name)), x) in
	    ini = pair(ini, i);
	}
	List::iter(init_one, inis);
	return Ksi::Bind(pair(ini, ref));
    }
}

Ksi::expr compile_union_expr(cg_ctx c, namedexpr u)
{
    let s = get_binding(u.nx_symbol) in {
	switch u.nx_expr.a.raw {
	case Void_const:  
	    return constant_union_expr(s);
	    
	case _:
	    push_proto(s);
	    let tag = symbol_union_tag(s) in
	    let t = type_of(u.nx_expr) in
	    let ex = s2t_conversion(t, compile_expr(c, u.nx_expr)) in
	    return call("gont_make_union", [tag, ex]);
	}
    }
}

Ksi::expr compile_assign(cg_ctx c, expr dst, expr src)
{
    switch dst.a.raw {
    case Tuple_expr(_):
	let n = make_id("tuple_assign$") in
	let v = make_vardef(tuple_ptr, n, compile_expr(c, src)) in
	let r = Ksi::Ref(n) in {
	    Ksi::expr assign = v;
	    
	    void walk(Ksi::expr handle, expr d)
	    {
		switch d.a.raw {
		case Tuple_expr(lst):
		    int pos = 0;
		    List::iter(fun void (expr x) 
				    {walk(tuple_nth(handle, pos++), x);},
				lst);
		case _:
		    assign = pair(assign,
				  Ksi::Modify(compile_expr(c, d), handle));
		}
	    }

	    walk(r, dst);

	    return Ksi::Bind(assign);
	}

    case _:
	_ need_save = false;
	switch src.a.raw {
	case Simple_binop(_, left, _):
	    switch left.a.raw {
	    case Assign_target: need_save = true;
	    case _:
	    }
	case _:
	}
	
	def compile_with_saved(target) 
	{
	    let src_ctx = c with {assign_target = Some(target)} in
	    let src = compile_expr(src_ctx, src) in
	    return Ksi::Modify(target, src);
	}
	
	if (need_save)
	    return maybe_save(c, dst, compile_with_saved);
	else
	    return Ksi::Modify(compile_expr(c, dst), compile_expr(c, src));
    }
}

Ksi::expr compile_dot_expr(cg_ctx c, expr ex, sf_symbol field)
{
    Ksi::id field_name = make_id(field.name);

    switch instanced_type_of(ex) {
    case Opt_struct(_):
	Ksi::expr make_check(Ksi::expr ptr)
	{
	    let check = Ksi::Cond(bin(Ksi::Eq, ptr, null_value), 
				  call("gont_null_access", []),
				  nop) in
	    return pair(check, Ksi::Arrow(ptr, field_name));
	}
	
	return maybe_save(c, ex, make_check);
    
    case Struct(_):
	return Ksi::Arrow(compile_expr(c, ex), field_name);

    case _:
	return ice("Codegen::compile_dot_expr(): non struct type");
    }
}

Ksi::expr compile_nl_expr(cg_ctx c, nl_expr nle)
{
    switch nle.raw {
    case Assign(e1, e2):
	return compile_assign(c, e1, e2);

    case Assign_target:
	switch c.assign_target {
	case Some(e):
	    return e;
	case None:
	    return ice("Codegen::compile_nl_expr(): assign target");
	}
    
    case Simple_binop(op, e1, e2):
	_ x1 = compile_expr(c, e1);
	_ x2 = compile_expr(c, e2);
	_ wrap = None;
	_ r;

	switch e1.a.raw {
	case Assign_target:
	    // it's saved anyway, and has_side_effects() ICEs on it
	    skip;
	    
	case _:
	    if (has_side_effects(e1) && has_side_effects(e2)) {
		let id = make_id(genname("bin_save"));
		wrap = Some(make_vardef(storage_type(type_of(e1)), id, x1));
		x1 = Ksi::Ref(id);
	    }
	}
	
	switch (instanced_type_of(e1), op) {
	case (String, Plus):
	    r = call("gont_strjoin", [x1, x2]);
	    
	case (String, Eq):
	    r = Ksi::Convert(Ksi::Bool, call("gont_streq", [x1, x2]));
	    
	case (String, Ne):
	    r = bin(Ksi::Eq, call("gont_streq", [x1, x2]), zero_const);
	
	case (String, _):
	    r = bin(ksi_op2(op), call("gont_strcmp", [x1, x2]), zero_const);
	
	case (Float, Div):
	    r = bin(Ksi::R_div, x1, x2);
	
	case (_, _):
	    r = bin(ksi_op2(op), x1, x2);
	}

	switch wrap {
	case Some(ex):
	    return pair(ex, r);
	case None:  
	    return r;
	}
	
    case Simple_unop(op, e):
	let x = compile_expr(c, e) in
	switch op {
	case Negate:	return Ksi::Unary(Ksi::Negate, x);
	case Bit_not:	return Ksi::Unary(Ksi::Bit_not, x);
	case Pre_dec:	return Ksi::Binary(Ksi::Pre_dec, x, one_const);
	case Pre_inc:	return Ksi::Binary(Ksi::Pre_inc, x, one_const);
	case Post_dec:	return Ksi::Binary(Ksi::Post_dec, x, one_const);
	case Post_inc:	return Ksi::Binary(Ksi::Post_inc, x, one_const);
	case Not:
	    // *don't* use Truth_not, it's broken
	    return Ksi::Binary(Ksi::Eq, Ksi::Convert(ptroff, x), zero_const);
	}
    
    case Ref(rr):
	return compile_ref(c, rr);

    case Call(e, el):
	let our_type = type_of({loc = top_location(), a = nle}) in
	return compile_call(c, our_type, e, el);

    case Tuple_expr(el):
	return compile_tuple(c, el);

    case Union_expr(u):
	return compile_union_expr(c, u);
    
    case Lambda(l):
	return compile_lambda(c, l);

    case Struct_expr(s, _):
	let our_type = type_of({loc = top_location(), a = nle}) in
	return compile_struct_expr(c, our_type, s, None);

    case Struct_copy(e, s, _):
	let our_type = type_of({loc = top_location(), a = nle}) in
	let ex = compile_expr(c, e) in
	return compile_struct_expr(c, our_type, s, Some(ex));

    case Dot_expr(de, _):
	return compile_dot_expr(c, de.sx_expr, get_binding(de.sx_symbol));
    
    case Bool_const(x):
	return Ksi::Bool_const(x);
    
    case Int_const(x):
	return Ksi::Int_const(storage_type(Int), x);
    
    case Float_const(x):
	return Ksi::Float_const(storage_type(Float), x);
	
    case String_const(x):
	let (_, len) = Iconv::string_to_utf8(x);
	let len_expr = Ksi::Int_const(ptroff, Int64::of_int(len));
	return call("gont_string_of_utf8", [Ksi::String_const(x), len_expr]);
	
    case Void_const:
	return null_value;
	
    case Null_const:
	return null_value;
    }
}

Ksi::expr compile_expr(cg_ctx c, expr e)
{
    return wfl_locate(c, compile_nl_expr, e);
}

void push_fn_decl(cg_ctx c, fn_desc fd)
{
    <*(Ksi::ty, Ksi::id)>list ksi_parms()
    {
	<var_symbol>list parm_syms = fd.lambda.lambda_parms;
	int id = 1;
	
	*(Ksi::ty, Ksi::id) compile_parm(*(ty, <id>option) parm)
	{
	    switch parm {
	    case (_, Some(_)):
		var_symbol s;
		switch parm_syms {
		case Cons(hd, tl):
		    parm_syms = tl;
		    s = hd;
		case _:
		    return ice("Codegen::push_fn_decl(): not enough symbols");
		}
		
		let t = transport_type(s.stype) in
		if (has_property(s, Is_in_closure) || 
		    need_t2s_conversion(s.stype))
		    return (t, make_id(s.name + "$tmp"));
		else
		    return (t, make_id(s.name));

	    case (t, None):
		return (transport_type(t), make_id("unused$" + itoa(id++)));
	    }
	    
	}
	
	*(Ksi::ty, Ksi::id) closure_parm()
	{
	    let p = parent_closure_type_name(c) in
	    if (p != "")
		return (Ksi::Ptr(Ksi::Struct(make_id(p))), parent_closure_id);
	    else
		return (null_type, dummy_id);
	}
	
	let ft = get_ft_record(fd.lambda.lambda_type) in
	let parms = List::map(compile_parm, ft.parms) in
	if (fd.takes_closure_parm)
	    return Cons(closure_parm(), parms);
	else
	    return parms;
    }

    Ksi::expr convert_parm(var_symbol s)
    {
	if (has_property(s, Is_in_closure)) {
	    let tmp = Ksi::Ref(make_id(s.name + "$tmp")) in
	    return vardef_init(c, s, t2s_conversion(s.stype, tmp));
	} else {
	    if (need_t2s_conversion(s.stype)) {
		let tmp = Ksi::Ref(make_id(s.name + "$tmp")) in
		return vardef_init(c, s, t2s_conversion(s.stype, tmp));
	    } else {
		return nop;
	    }
	}
    }

    // definition of closure variable, allocation of memory for closure
    // and possibly assignment of parent_closure$ field.
    Ksi::expr closure_stuff()
    {
	if (!fd.lambda.lambda_needs_closure) {
	    // ha! no closure
	    c.cf_closure_type_name = "";
	    return nop;
	}
	
	push_closure_type(c, fd.lambda.lambda_closure_vars);
	
	Ksi::expr ret;
	
	let closure_type = Ksi::Struct(make_id(c.cf_closure_type_name)) in
	let clptr = Ksi::Ptr(closure_type) in
	let mem = Ksi::Convert(clptr, gont_alloc(Ksi::Size_of(closure_type))) in
	ret = make_vardef(clptr, our_closure_id, mem);

	if (parent_closure_type_name(c) != "") {
	    let par = Ksi::Arrow(Ksi::Ref(our_closure_id), parent_closure_id) in
	    let ini = Ksi::Modify(par, Ksi::Ref(parent_closure_id)) in
	    ret = pair(ret, ini);
	}
	    
	return ret;
    }
    
    c = {cf_is_public = fd.is_public,
	 cf_name = fd.name,
	 cf_ret_type_instanced = 
		type_instance((get_ft_record(fd.lambda.lambda_type)).ret_type),
	 parent = Some(c),
	 parent_func = Some(c),
	 cf_closure_type_name = "?",
	 cf_references_hash = Ptrhash::create(13),
	 cf_references = ref_make(Nil),
	 assign_target = None
    };
    
    
    if (fd.is_public)
        top_comment("public " + fd.name);
    else
	top_comment("local " + fd.name);
    
    _ ksi_ret_type;
    
    switch c.cf_ret_type_instanced {
    case Void:
	ksi_ret_type = Ksi::Void;
    case _:
	ksi_ret_type = transport_type(c.cf_ret_type_instanced);
    }
   
    let closure = closure_stuff() in
    let parms_init = List::map(convert_parm, fd.lambda.lambda_parms) in
    let startup = Ksi::Compound(Cons(closure, parms_init)) in {
	Ksi::expr body = compile_stmt(c, fd.lambda.lambda_body);
	
	let ref_lst = Cons(startup, ref_get(c.cf_references)) in
	body = Ksi::Bind(pair(Ksi::Compound(ref_lst), Ksi::Bind(body)));
	
	Ksi::decl d = Ksi::Func(public_attrs(fd.is_public), 
				ksi_ret_type, 
				make_id(fd.name),
				ksi_parms(), 
				body);
				
	top_comment("body " + fd.name);
	output_decl(d);
	top_comment("end " + fd.name);
    }
}

Ksi::expr compile_lambda(cg_ctx c, lambda_record l)
{
    string name = genname(c.cf_name + "$l");
    
    push_fn_decl(c, ({name = name,
		      is_public = false,
		      takes_closure_parm = true,
		      lambda = l}));
   
    if (c.cf_closure_type_name == "")
	return static_lambda(name);
    else 
	return call("gont_make_lambda", [Ksi::Func_ref(make_id(name)), 
					 Ksi::Ref(our_closure_id)]);
}

// vim:sts=4:sw=4
