/*
 * Copyright (c) 2002 Michal Moskal <malekith@pld-linux.org>
 * All rights reserved. See file COPYRIGHT for details about licensing.
 *
 * $Rev: 840 $ $LastChangedDate: 2002-09-01 16:56:25 +0200 (Sun, 01 Sep 2002) $
 *
 * Tyutil - Utilities for type manipulation.
 */

open Ast;

/* Return string representation of [t], for example for error messages. 
   Safe to call on cyclic types.  */
string string_of_type(ty t);

/* Return [t] with all levels of X_tyvar(...) indirection stripped off.
   It is safe to call on cyclic type in which case it will report error
   message and return Abstract_type.  */
ty normalize_xt(ty t);

/* Strip all X_tyvar() stuff until some real type.  ICE if None if found.  
   To be used by passes after typecheck, which can generally assume there
   are no weak types in what they operate on.  */
ty strip_xt(ty t);

/* Substitute X_tyvar(...) in type with their content.  Often good for error 
   messages, but not always.  */
ty clean_type(ty t);

/* Substitute fresh type variables in [t] for type variables listed in [ns].
   Return list of variables substituted and substituted type.  Probably there
   is no need to use this beside Type and Tyutil modules.  */
*(<xtyvar>list, ty) subst_xtyvars(ty t, <id>list ns);

/* For each pair [(name, type)] in [assoc] substitute all occurrences of
   [Tyvar(name)] with [type] in [t].  Return substituted [t].  Probably there
   is no need to use this beside Type and Tyutil modules.  */
ty subst_vars(ty t, <*(id, ty)>list assoc);

/* Strip one level of Named_type(...) indirection.  Type arguments are 
   substituted as necessary.  Will ICE on weak types.  */
ty type_instance_one(ty t);

/* Test whether [t] is base type, i.e. is not just an alias for some other 
   type.  Will ICE on weak types.  */
bool is_base_type(ty t);

/* Test whether [t] is fully instanced type, i.e. is not Named_type(...).  
   Will ICE on weak types.  */
bool is_type_instance(ty t);

/* Return real content of type, i.e. with all Named_type(...) stripped off.
   It will return Struct(...), Opt_struct(...) and Union(...) as appropriate.
   Won't ever return Named_type(...) or X_tyvar(...).  Will ICE on weak types.  */
ty type_instance(ty t);

/* Return copy of [tt] with all type aliases expanded to their real meaning.
   However Named_types(...) for structures and unions are not expanded.
   This functions just expands toplevel [t], not its content.  See 
   canonical_type() for recursive version. Will ICE on weak types.  */
ty base_type(ty t);

/* Same as base_type() except it traverses entire type tree expanding
   Named_types(...).  */
ty canonical_type(ty tt);

/* Return type of [e].  Can only be called after expression has been type 
   checked.  */
ty type_of(expr e);

/* Return fully expanded type of [e].  Can only be called after expression
   has been type checked.  */
ty instanced_type_of(expr e);

/* If [t] is an function type, return its [ft_record].  Otherwise ICE.  */
ft_record get_ft_record(ty t);

/* Return [true] iff [t] is type of function returning Void.  */
bool is_void_func_type(ty t);

/* Generate new weak type variable, name of which will start with [name].  */
xtyvar genxt(string name);

/* Return string being flat terse representation of [t], using 
   [[A-Za-z_0-9:<>]] characters only.  It's used for debugging information.  */
string mangle_type(ty t);

/* Lookup union of given one of labels contained in it.  */
ty_symbol union_of(var_symbol label);

/* Given a structure type, lookup named field in it.  Print error message
   and raise Choke if not found.  */
<sf_symbol>field find_field_in_struct(ty t, <sf_symbol>bindable name);

/* Given name of field. lookup its type in given union/struct type.  */
ty field_type_in(ty t, <<'a>symbol>bindable name);

/* Check if [e] is valid L-value, i.e. can be used as target of assignment 
   expression.  */
bool is_lvalue(expr e);

/* Check whatever given expression is constant, i.e. can be evaluated at
   compile-time.  This is kinds conservative.  */
bool is_constant_expr(expr e);

/* Check whatever given symbol is function defined in global scope, which
   means it has different calling convention -- no closure.  */
bool is_global_function(var_symbol s);

/* Check if there are none weak type variables in [t].  If they are, issue
   error message and return [true], otherwise return [false].  */
bool check_weak(ty t);

/* Return conservative estimation whether [ex] might have any side effects.
   Allocation is not counted as a side-effect.  */
bool has_side_effects(expr ex);

/* Return [true] iff [t] contains any type variables.  */
bool is_poly(ty t);

// vim:sts=4:sw=4
